#!/bin/sh
#
# Copyright (c) 2006-2021 Varnish Software AS
# SPDX-License-Identifier: BSD-2-Clause
# See LICENSE file for full text of license

FLOPS='
	*.c
	../../lib/libvinylapi/flint.lnt
	../../lib/libvinylapi/*.c
' ../../tools/flint_skel.sh $*
